package ru.dbotthepony.kommons.util;

import java.security.DigestException;
import java.security.MessageDigest;
import java.util.Objects;

/**
 * Pure Java implementation of XXHash64 hashing function.
 * <p>
 * {@link XXHash64#digest()} returns Long encoded as big-endian number.
 */
public final class XXHash64 extends MessageDigest {
	private final long seed;
	private final byte[] mem = new byte[32];
	private int memIndex;

	private long length;

	private long v0;
	private long v1;
	private long v2;
	private long v3;

	public static final long XXH_PRIME64_1 = 0x9E3779B185EBCA87L;  /*!< 0b1001111000110111011110011011000110000101111010111100101010000111 */
	public static final long XXH_PRIME64_2 = 0xC2B2AE3D27D4EB4FL;  /*!< 0b1100001010110010101011100011110100100111110101001110101101001111 */
	public static final long XXH_PRIME64_3 = 0x165667B19E3779F9L;  /*!< 0b0001011001010110011001111011000110011110001101110111100111111001 */
	public static final long XXH_PRIME64_4 = 0x85EBCA77C2B2AE63L;  /*!< 0b1000010111101011110010100111011111000010101100101010111001100011 */
	public static final long XXH_PRIME64_5 = 0x27D4EB2F165667C5L;  /*!< 0b0010011111010100111010110010111100010110010101100110011111000101 */

	public XXHash64(long seed) {
		super("XXHash64");
		this.seed = seed;
		engineReset();
	}

	public XXHash64() {
		this(0);
	}

	private long load64(int index) {
		return load64(mem, index);
	}

	private long load64(byte[] source, int index) {
		return (((long) source[index]) & 0xFFL) |
			(((long) source[index + 1]) & 0xFFL) << 8 |
			(((long) source[index + 2]) & 0xFFL) << 16 |
			(((long) source[index + 3]) & 0xFFL) << 24 |
			(((long) source[index + 4]) & 0xFFL) << 32 |
			(((long) source[index + 5]) & 0xFFL) << 40 |
			(((long) source[index + 6]) & 0xFFL) << 48 |
			(((long) source[index + 7]) & 0xFFL) << 56;
	}

	private long load32(int index) {
		return (((long) mem[index]) & 0xFFL) |
			(((long) mem[index + 1]) & 0xFFL) << 8 |
			(((long) mem[index + 2]) & 0xFFL) << 16 |
			(((long) mem[index + 3]) & 0xFFL) << 24;
	}

	private long round(long acc, long input) {
		acc += input * XXH_PRIME64_2;
		acc = Long.rotateLeft(acc, 31);
		return acc * XXH_PRIME64_1;
	}

	private long mergeRound(long acc, long val) {
		val = round(0L, val);
		acc ^= val;
		return acc * XXH_PRIME64_1 + XXH_PRIME64_4;
	}

	private void doRound() {
		memIndex = 0;
		v0 = round(v0, load64(0));
		v1 = round(v1, load64(8));
		v2 = round(v2, load64(16));
		v3 = round(v3, load64(24));
	}

	@Override
	protected void engineUpdate(byte input) {
		length++;
		mem[memIndex++] = input;

		if (memIndex >= 32)
			doRound();
	}

	@Override
	protected void engineUpdate(byte[] input, int offset, int len) {
		Objects.checkFromIndexSize(offset, input.length, len);
		if (len == 0) return;

		length += len;

		if (memIndex != 0) {
			int size = Math.min(32 - memIndex, len);
			System.arraycopy(input, offset, mem, memIndex, size);
			memIndex += size;
			offset += size;
			len -= size;

			if (memIndex >= 32)
				doRound();
		}

		while (len >= 32) {
			v0 = round(v0, load64(input, offset));
			v1 = round(v1, load64(input, offset + 8));
			v2 = round(v2, load64(input, offset + 16));
			v3 = round(v3, load64(input, offset + 24));
			offset += 32;
			len -= 32;
		}

		if (len != 0) {
			System.arraycopy(input, offset, mem, memIndex, len);
			memIndex += len;

			if (memIndex >= 32)
				doRound();
		}
	}

	public long digestAsLong() {
		long h64;

		if (length >= 32L) {
			h64 = Long.rotateLeft(v0, 1) + Long.rotateLeft(v1, 7) + Long.rotateLeft(v2, 12) + Long.rotateLeft(v3, 18);
			h64 = mergeRound(h64, v0);
			h64 = mergeRound(h64, v1);
			h64 = mergeRound(h64, v2);
			h64 = mergeRound(h64, v3);
		} else {
			h64 = seed + XXH_PRIME64_5;
		}

		h64 += length;
		int len = (int) (length & 31L);
		int index = 0;

		while (len >= 8) {
			h64 ^= round(0L, load64(index));
			h64 = Long.rotateLeft(h64, 27) * XXH_PRIME64_1 + XXH_PRIME64_4;
			index += 8;
			len -= 8;
		}

		if (len >= 4) {
			h64 ^= load32(index) * XXH_PRIME64_1;
			h64 = Long.rotateLeft(h64, 23) * XXH_PRIME64_2 + XXH_PRIME64_3;
			index += 4;
			len -= 4;
		}

		while (len-- > 0) {
			h64 ^= (mem[index++] & 0xFFL) * XXH_PRIME64_5;
			h64 = Long.rotateLeft(h64, 11) * XXH_PRIME64_1;
		}

		h64 ^= h64 >>> 33;
		h64 *= XXH_PRIME64_2;
		h64 ^= h64 >>> 29;
		h64 *= XXH_PRIME64_3;
		h64 ^= h64 >>> 32;

		engineReset();

		return h64;
	}

	@Override
	protected int engineGetDigestLength() {
		return 8;
	}

	@Override
	protected byte[] engineDigest() {
		long h64 = digestAsLong();

		byte[] result = new byte[8];

		result[7] = (byte) h64;
		result[6] = (byte) (h64 >>> 8);
		result[5] = (byte) (h64 >>> 16);
		result[4] = (byte) (h64 >>> 24);
		result[3] = (byte) (h64 >>> 32);
		result[2] = (byte) (h64 >>> 40);
		result[1] = (byte) (h64 >>> 48);
		result[0] = (byte) (h64 >>> 56);

		return result;
	}

	@Override
	protected int engineDigest(byte[] buf, int offset, int len) throws DigestException {
		if (len < 8)
			throw new DigestException("partial digests not returned");

		if (buf.length - offset < 8)
			throw new DigestException("insufficient space in the output "
				+ "buffer to store the digest");

		long h64 = digestAsLong();

		buf[offset + 7] = (byte) h64;
		buf[offset + 6] = (byte) (h64 >>> 8);
		buf[offset + 5] = (byte) (h64 >>> 16);
		buf[offset + 4] = (byte) (h64 >>> 24);
		buf[offset + 3] = (byte) (h64 >>> 32);
		buf[offset + 2] = (byte) (h64 >>> 40);
		buf[offset + 1] = (byte) (h64 >>> 48);
		buf[offset] = (byte) (h64 >>> 56);

		return 8;
	}

	@Override
	protected void engineReset() {
		this.length = 0L;
		this.memIndex = 0;
		this.v0 = this.seed + XXH_PRIME64_1 + XXH_PRIME64_2;
		this.v1 = this.seed + XXH_PRIME64_2;
		this.v2 = this.seed;
		this.v3 = this.seed - XXH_PRIME64_1;
	}
}
