package ru.dbotthepony.kommons.io

import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import ru.dbotthepony.kommons.collect.ListenableMap
import ru.dbotthepony.kommons.collect.ListenableSet
import ru.dbotthepony.kommons.math.Decimal
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kommons.util.DelegateGetter
import ru.dbotthepony.kommons.util.DelegateSetter
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.util.Listenable
import ru.dbotthepony.kommons.util.ListenableDelegate
import ru.dbotthepony.kommons.util.Observer
import ru.dbotthepony.kommons.util.ValueObserver
import java.io.Closeable
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.util.*
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.collections.ArrayList
import kotlin.concurrent.withLock

/**
 * Universal, one-to-many delegate/value synchronizer.
 *
 * Values and delegates can be attached using [add], or by constructing subclassed directly.
 * Delta changes are tracked by [DelegateSyncher.Remote] instances.
 *
 * In general, this class is not meant to be concurrently mutated by different threads,
 * to avoid structure corruption a lock is employed.
 *
 * Attached delegates can be safely mutated by multiple threads concurrently
 * (attached [ListenableDelegate] can call [Listenable.addListener] callback concurrently).
 */
class DelegateSyncher : Observer {
	private val lock = ReentrantLock()
	private val slots = ArrayList<AbstractSlot?>()
	private val gaps = IntAVLTreeSet()
	private val observers = ArrayList<AbstractSlot>()
	private val remotes = ArrayList<Remote>()
	private var isRemote = false

	override fun observe(): Boolean {
		var any = false
		observers.forEach { any = it.observe() || any }
		return any
	}

	fun read(stream: DataInputStream) {
		isRemote = true

		var readID = stream.readVarInt()

		while (readID > 0) {
			val slot = slots.getOrNull(readID - 1) ?: throw IndexOutOfBoundsException("Unknown networked slot ${readID - 1}!")
			slot.read(stream)
			readID = stream.readVarInt()
		}
	}

	fun read(stream: InputStream) {
		if (stream is DataInputStream)
			return read(stream)

		read(DataInputStream(stream))
	}

	abstract inner class AbstractSlot : Closeable, Observer, Comparable<AbstractSlot> {
		val id: Int

		protected val isRemoved = AtomicBoolean()

		fun remove() {
			if (isRemoved.compareAndSet(false, true)) {
				lock.withLock {
					slots[id] = null
					gaps.add(id)

					remoteSlots.forEach {
						it.remove()
					}

					remoteSlots.clear()
					remove0()
				}
			}
		}

		final override fun close() {
			return remove()
		}

		final override fun compareTo(other: AbstractSlot): Int {
			return id.compareTo(other.id)
		}

		protected abstract fun remove0()
		internal abstract fun read(stream: DataInputStream)
		internal abstract fun write(stream: DataOutputStream)

		internal val remoteSlots = CopyOnWriteArrayList<Remote.RemoteSlot>()

		protected fun markDirty() {
			if (!isRemote && !isRemoved.get()) {
				remoteSlots.forEach {
					it.markDirty()
				}
			}
		}

		init {
			lock.withLock {
				if (gaps.isNotEmpty()) {
					val gap = gaps.firstInt()
					gaps.remove(gap)
					id = gap
					check(slots[id] == null) { "Expected slot $id to be empty" }
					slots[id] = this
				} else {
					id = slots.size
					slots.add(this)
				}

				remotes.forEach {
					it.addSlot(this)
				}
			}
		}
	}

	inner class Slot<V>(val delegate: ListenableDelegate<V>, val codec: StreamCodec<V>) : AbstractSlot(), ListenableDelegate<V> {
		private val l = delegate.addListener(Runnable {
			markDirty()
		})

		init {
			if (delegate is ListenableDelegate.AbstractShadow) {
				observers.add(this)
			}
		}

		private val listeners = Listenable.Impl<V>()

		override fun remove0() {
			l.remove()
			listeners.clear()
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			delegate.accept(codec.read(stream))
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			codec.write(stream, delegate.get())
		}

		override fun get(): V {
			check(!isRemoved.get()) { "This network slot was removed" }
			return delegate.get()
		}

		override fun accept(t: V) {
			check(!isRemoved.get()) { "This network slot was removed" }
			delegate.accept(t)
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return listeners.addListener(listener)
		}

		override fun observe(): Boolean {
			check(!isRemoved.get()) { "This network slot was removed" }

			if (delegate is ListenableDelegate.AbstractShadow) {
				return delegate.observe()
			}

			return false
		}
	}

	inner class ObservedSlot<V>(val delegate: Delegate<V>, val codec: StreamCodec<V>) : AbstractSlot(), ListenableDelegate<V>, ValueObserver<V> {
		private val listeners = Listenable.Impl<V>()
		private var observed: Any? = Mark
		private val observeLock = ReentrantLock()

		init {
			observers.add(this)
		}

		override fun remove0() {
			listeners.clear()
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			val read = codec.read(stream)
			observed = codec.copy(read)
			delegate.accept(read)
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			codec.write(stream, delegate.get())
		}

		override fun observe(): Boolean {
			val get = delegate.get()

			if (observed === Mark || !codec.compare(get, observed as V)) {
				observeLock.withLock {
					if (observed === Mark || !codec.compare(get, observed as V)) {
						observed = codec.copy(get)
						markDirty()
						listeners.accept(get)
					}

					return true
				}
			}

			return false
		}

		override fun getAndObserve(): Pair<V, Boolean> {
			val get = delegate.get()

			if (observed === Mark || !codec.compare(get, observed as V)) {
				observeLock.withLock {
					if (observed === Mark || !codec.compare(get, observed as V)) {
						observed = codec.copy(get)
						markDirty()
						listeners.accept(get)
					}

					return get to  true
				}
			}

			return get to false
		}

		override fun get(): V {
			check(!isRemoved.get()) { "This network slot was removed" }
			return delegate.get()
		}

		override fun accept(t: V) {
			check(!isRemoved.get()) { "This network slot was removed" }
			delegate.accept(t)
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return listeners.addListener(listener)
		}
	}

	private object Mark

	internal data class SetAction<V>(val value: KOptional<V>, val action: Int)

	inner class SetSlot<V>(val delegate: ListenableSet<V>, val codec: StreamCodec<V>) : AbstractSlot() {
		private val listener = object : ListenableSet.SetListener<V> {
			override fun onClear() {
				remoteSlots.forEach {
					(it as Remote.RemoteSetSlot<V>).changelist.clear()

					synchronized(it.changelistLock) {
						it.changelist.add(SetAction(KOptional(), CLEAR))
					}

					it.markDirty()
				}
			}

			override fun onValueAdded(element: V) {
				remoteSlots.forEach {
					it as Remote.RemoteSetSlot<V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.value.valueEquals(element) }
						it.changelist.add(SetAction(KOptional(codec.copy(element)), ADD))
					}

					it.markDirty()
				}
			}

			override fun onValueRemoved(element: V) {
				remoteSlots.forEach {
					it as Remote.RemoteSetSlot<V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.value.valueEquals(element) }
						it.changelist.add(SetAction(KOptional(codec.copy(element)), REMOVE))
					}

					it.markDirty()
				}
			}
		}

		private val l = delegate.addListener(listener)

		override fun remove0() {
			l.remove()
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			var action = stream.read()

			while (true) {
				when (action) {
					ADD -> delegate.add(codec.read(stream))
					REMOVE -> delegate.remove(codec.read(stream))
					CLEAR -> delegate.clear()
					else -> break
				}

				action = stream.read()
			}
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			throw RuntimeException("unreachable code")
		}

		override fun observe(): Boolean {
			check(!isRemoved.get()) { "This network slot was removed" }
			return false
		}
	}

	internal data class MapAction<K, V>(val key: KOptional<K>, val value: KOptional<V>, val action: Int)

	inner class MapSlot<K, V>(val delegate: ListenableMap<K, V>, val keyCodec: StreamCodec<K>, val valueCodec: StreamCodec<V>) : AbstractSlot() {
		private val listener = object : ListenableMap.MapListener<K, V> {
			override fun onClear() {
				remoteSlots.forEach {
					it as Remote.RemoteMapSlot<K, V>

					synchronized(it.changelistLock) {
						it.changelist.clear()
						it.changelist.add(MapAction(KOptional(), KOptional(), CLEAR))
					}

					it.markDirty()
				}
			}

			override fun onValueAdded(key: K, value: V) {
				remoteSlots.forEach {
					it as Remote.RemoteMapSlot<K, V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.key.valueEquals(key) }
						it.changelist.add(MapAction(KOptional(keyCodec.copy(key)), KOptional(valueCodec.copy(value)), ADD))
					}

					it.markDirty()
				}
			}

			override fun onValueRemoved(key: K, value: V) {
				remoteSlots.forEach {
					it as Remote.RemoteMapSlot<K, V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.key.valueEquals(key) }
						it.changelist.add(MapAction(KOptional(keyCodec.copy(key)), KOptional(valueCodec.copy(value)), REMOVE))
					}

					it.markDirty()
				}
			}
		}

		private val l = delegate.addListener(listener)

		override fun remove0() {
			l.remove()
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			var action = stream.read()

			while (true) {
				when (action) {
					ADD -> delegate.put(keyCodec.read(stream), valueCodec.read(stream))
					REMOVE -> delegate.remove(keyCodec.read(stream))
					CLEAR -> delegate.clear()
					else -> break
				}

				action = stream.read()
			}
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved.get()) { "This network slot was removed" }
			throw RuntimeException("unreachable code")
		}

		override fun observe(): Boolean {
			check(!isRemoved.get()) { "This network slot was removed" }
			return false
		}
	}

	fun <V> add(delegate: ListenableDelegate<V>, codec: StreamCodec<V>): Slot<V> {
		return Slot(delegate, codec)
	}

	fun <V> add(delegate: Delegate<V>, codec: StreamCodec<V>): ObservedSlot<V> {
		return ObservedSlot(delegate, codec)
	}

	fun <V> add(delegate: Supplier<V>, codec: StreamCodec<V>): Slot<V> {
		return Slot(ListenableDelegate.CustomShadow(delegate, codec::compare, codec::copy), codec)
	}

	fun <E> add(delegate: ListenableSet<E>, codec: StreamCodec<E>): SetSlot<E> {
		return SetSlot(delegate, codec)
	}

	fun <K, V> add(delegate: ListenableMap<K, V>, keyCodec: StreamCodec<K>, valueCodec: StreamCodec<V>): MapSlot<K, V> {
		return MapSlot(delegate, keyCodec, valueCodec)
	}

	@JvmName("vbyte")
	@JvmOverloads
	fun byte(value: Byte = 0, setter: DelegateSetter<Byte> = DelegateSetter.passthrough(), getter: DelegateGetter<Byte> = DelegateGetter.passthrough()): Slot<Byte> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), ByteValueCodec)
	}

	@JvmName("vshort")
	@JvmOverloads
	fun short(value: Short = 0, setter: DelegateSetter<Short> = DelegateSetter.passthrough(), getter: DelegateGetter<Short> = DelegateGetter.passthrough()): Slot<Short> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), ShortValueCodec)
	}

	@JvmName("vchar")
	@JvmOverloads
	fun char(value: Char, setter: DelegateSetter<Char> = DelegateSetter.passthrough(), getter: DelegateGetter<Char> = DelegateGetter.passthrough()): Slot<Char> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), CharValueCodec)
	}

	@JvmName("vint")
	@JvmOverloads
	fun int(value: Int = 0, setter: DelegateSetter<Int> = DelegateSetter.passthrough(), getter: DelegateGetter<Int> = DelegateGetter.passthrough()): Slot<Int> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), VarIntValueCodec)
	}

	@JvmName("vlong")
	@JvmOverloads
	fun long(value: Long = 0L, setter: DelegateSetter<Long> = DelegateSetter.passthrough(), getter: DelegateGetter<Long> = DelegateGetter.passthrough()): Slot<Long> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), VarLongValueCodec)
	}

	@JvmName("vfloat")
	@JvmOverloads
	fun float(value: Float = 0f, setter: DelegateSetter<Float> = DelegateSetter.passthrough(), getter: DelegateGetter<Float> = DelegateGetter.passthrough()): Slot<Float> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), FloatValueCodec)
	}

	@JvmName("vdouble")
	@JvmOverloads
	fun double(value: Double = 0.0, setter: DelegateSetter<Double> = DelegateSetter.passthrough(), getter: DelegateGetter<Double> = DelegateGetter.passthrough()): Slot<Double> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), DoubleValueCodec)
	}

	@JvmName("vboolean")
	@JvmOverloads
	fun boolean(value: Boolean = false, setter: DelegateSetter<Boolean> = DelegateSetter.passthrough(), getter: DelegateGetter<Boolean> = DelegateGetter.passthrough()): Slot<Boolean> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), BooleanValueCodec)
	}

	@JvmOverloads
	fun string(value: String, setter: DelegateSetter<String> = DelegateSetter.passthrough(), getter: DelegateGetter<String> = DelegateGetter.passthrough()): Slot<String> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), BinaryStringCodec)
	}

	@JvmOverloads
	fun uuid(value: UUID, setter: DelegateSetter<UUID> = DelegateSetter.passthrough(), getter: DelegateGetter<UUID> = DelegateGetter.passthrough()): Slot<UUID> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), UUIDValueCodec)
	}

	@JvmOverloads
	fun decimal(value: Decimal = Decimal.ZERO, setter: DelegateSetter<Decimal> = DelegateSetter.passthrough(), getter: DelegateGetter<Decimal> = DelegateGetter.passthrough()): Slot<Decimal> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), DecimalValueCodec)
	}

	fun <E> set(codec: StreamCodec<E>, backing: MutableSet<E> = ObjectOpenHashSet()): SetSlot<E> {
		return add(ListenableSet(backing), codec)
	}

	fun <K, V> map(keyCodec: StreamCodec<K>, valueCodec: StreamCodec<V>, backing: MutableMap<K, V> = Object2ObjectOpenHashMap()): MapSlot<K, V> {
		return add(ListenableMap(backing), keyCodec, valueCodec)
	}

	/**
	 * Remotes must be [close]d when no longer in use, otherwise they will
	 * leak memory and decrease performance of syncher.
	 *
	 * To get changes to be networked to remote client, [write] should be called.
	 */
	inner class Remote : Closeable {
		private val isRemoved = AtomicBoolean()
		internal val dirty = ConcurrentLinkedQueue<RemoteSlot>()
		private val remoteSlots = CopyOnWriteArrayList<RemoteSlot>()

		internal open inner class RemoteSlot(val parent: AbstractSlot) : Comparable<RemoteSlot> {
			private val isDirty = AtomicBoolean()

			final override fun compareTo(other: RemoteSlot): Int {
				return parent.compareTo(other.parent)
			}

			init {
				remoteSlots.add(this)
				dirty.add(this)
			}

			fun markDirty() {
				if (!isRemoved.get() && isDirty.compareAndSet(false, true)) {
					dirty.add(this)
				}
			}

			fun markClean() {
				isDirty.set(false)
			}

			open fun invalidate() {
				markDirty()
			}

			open fun remove() {
				isDirty.set(true)
				remoteSlots.remove(this)
				dirty.remove(this)
			}

			open fun write(stream: DataOutputStream) {
				parent.write(stream)
			}

			override fun hashCode(): Int {
				return parent.hashCode()
			}

			override fun equals(other: Any?): Boolean {
				return this === other || other is RemoteSlot && other.parent == parent
			}
		}

		internal inner class RemoteSetSlot<V>(parent: SetSlot<V>) : RemoteSlot(parent) {
			val changelist = ArrayList<SetAction<V>>()
			val changelistLock = Any()

			override fun remove() {
				super.remove()
				changelist.clear()
			}

			override fun invalidate() {
				synchronized(changelistLock) {
					changelist.clear()
					changelist.add(SetAction(KOptional(), CLEAR))

					parent as SetSlot<V>

					for (v in parent.delegate) {
						changelist.add(SetAction(KOptional(parent.codec.copy(v)), ADD))
					}
				}

				super.invalidate()
			}

			override fun write(stream: DataOutputStream) {
				synchronized(changelistLock) {
					for (change in changelist) {
						stream.write(change.action)
						change.value.ifPresent { (parent as SetSlot<V>).codec.write(stream, it) }
					}

					changelist.clear()
				}

				stream.write(0)
			}
		}

		internal inner class RemoteMapSlot<K, V>(parent: MapSlot<K, V>) : RemoteSlot(parent) {
			val changelist = ArrayList<MapAction<K, V>>()
			val changelistLock = Any()

			override fun remove() {
				super.remove()
				changelist.clear()
			}

			override fun invalidate() {
				synchronized(changelistLock) {
					changelist.clear()
					changelist.add(MapAction(KOptional(), KOptional(), CLEAR))

					parent as MapSlot<K, V>

					for ((k, v) in parent.delegate) {
						changelist.add(MapAction(KOptional(parent.keyCodec.copy(k)), KOptional(parent.valueCodec.copy(v)), ADD))
					}
				}

				super.invalidate()
			}

			override fun write(stream: DataOutputStream) {
				synchronized(changelistLock) {
					for (change in changelist) {
						stream.write(change.action)
						change.key.ifPresent { (parent as MapSlot<K, V>).keyCodec.write(stream, it) }
						change.value.ifPresent { (parent as MapSlot<K, V>).valueCodec.write(stream, it) }
					}

					changelist.clear()
				}

				stream.write(0)
			}
		}

		init {
			lock.withLock {
				slots.forEach {
					if (it != null) {
						addSlot(it)
					}
				}

				remotes.add(this)
			}
		}

		internal fun addSlot(slot: AbstractSlot) {
			if (slot is SetSlot<*>) {
				slot.remoteSlots.add(RemoteSetSlot(slot))
			} else if (slot is MapSlot<*, *>) {
				slot.remoteSlots.add(RemoteMapSlot(slot))
			} else {
				slot.remoteSlots.add(RemoteSlot(slot))
			}
		}

		/**
		 * Returns null of this remote is clean.
		 */
		fun write(): FastByteArrayOutputStream? {
			this@DelegateSyncher.observe()

			if (dirty.isNotEmpty()) {
				val data = FastByteArrayOutputStream()
				val stream = DataOutputStream(data)

				val sorted = ObjectAVLTreeSet<RemoteSlot>()
				var next = dirty.poll()

				while (next != null) {
					sorted.add(next)
					next.markClean()
					next = dirty.poll()
				}

				for (v in sorted) {
					stream.writeVarInt(v.parent.id + 1)
					v.write(stream)
				}

				stream.write(0)

				return data
			}

			return null
		}

		/**
		 * Marks all networked slots dirty
		 */
		fun invalidate() {
			remoteSlots.forEach { it.invalidate() }
		}

		override fun close() {
			if (isRemoved.compareAndSet(false, true)) {
				lock.withLock {
					remoteSlots.forEach {
						it.remove()
						it.parent.remoteSlots.remove(it)
					}

					remotes.remove(this)
				}

				dirty.clear()
			}
		}
	}

	companion object {
		private const val END = 0
		private const val CLEAR = 1
		private const val ADD = 2
		private const val REMOVE = 3
	}
}
